#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"

enum {						/* CSR registers */
	Config		= 0x0000,
	Intcoalescing	= 0x0004,		/* accumulate interrupts, 32us units */
	Icr			= 0x0008,		/* Interrupt status/control */
	Intmask		= 0x000c,		/* Interrupt mask */
	Fhicr			= 0x0010,		/* FH interrupt status/control */
	Reset		= 0x0020,
	Gpctl		= 0x0024,
	Revision		= 0x0028,		/* revision bits: 7-4 type, 3-2 step, 1-0 dash */
	EEreg		= 0x002c,
	EEgp			= 0x0030,
	Gio			= 0x003c,
	Gp1clr		= 0x005c,
	Giobits		= 0x0100,
	Anlgpll		= 0x020c,		/* analog phase-lock-loop */
	Warevision	= 0x022c,		/* HW revision used for CCK backoff */
	Hpetmem		= 0x0240,
	Hbusraddr	= 0x040c,		/* select SRAM address to read from */
	Hbuswaddr	= 0x0410,		/* select SRAM address to write to */
	Hbuswdata	= 0x0418,		/* data to write to the SRAM address */
	Hbusrdata	= 0x041c,		/* data to read from the SRAM address */
	Prphwaddr	= 0x0444,		/* address of peripheral register to write to */
	Prphraddr		= 0x0448,		/* address of peripheral register to read from */
	Prphwdata	= 0x044c,		/* data to write */
	Prphrdata		= 0x0450,		/* data to read */
	Tqhwp		= 0x0460,		/* Tx queue hardware wp */
	Kwaddr		= 0x197c,		/* keep-warm buffer address */
	Tqbd		= 0x19d0,		/* Tx queue descriptor ring base at Tqbd + tid*4 */
	Rqst			= 0x1bc0,		/* Rx queue status address */
	Rqbd		= 0x1bc4,		/* Rx queue descriptor ring base */
	Rqhwp		= 0x1bc8,		/* Rx queue hardware wp */
	Tqconfig		= 0x1d00,		/* Tx config at Tqconfig + 0x20*dmachan */
	Rqconfig		= 0x1c00,		/* Rx queue config */
	Rqstatus		= 0x1c44,		/* Rx queue status (poll for bit 24 after setting DMA off */
	Tqstatus		= 0x1eb0,		/* Tx queue status */
	Tsvcconfig	= Tqconfig + 0x20*9,
	Tsvcsaddr		= 0x19c8,
	Tsvcf0		= 0x1900 + 8*9,
	Tsvcf1		= Tsvcf0 + 4,
	Tsvcsts		= 0x1d08 + 0x20*9,
	
};

enum {						/* PRPH registers */
	Apmgclkenable= 0x3004,
	Apmgclkdisable= 0x3008,
	Apmgpwr		= 0x300c,
	Apmgpcistt	= 0x3010,
	Tqfact		= 0xa02c10,	/* enable/disable DMA channel for each bit */
	Scdsaddr		= 0xa02c00,	/* SRAM address for device scheduler */
	Scddaddr		= 0xa02c08,	/* DRAM address for device scheduler */
	Scdqchain	= 0xa02ce8,	/* select which queues work in chain mode */
	Scdaggr		= 0xa02e48,
	Scdqhrp		= 0xa02c68,	/* hrp for the 0th queue */
	Scdqhwp		= 0xa02c18,	/* hwp for the 0th queue */
	Scdintrmask	= 0xa02d08,	/* interrupt mask (which queue causes interrupt) */
	Scdqstatus	= 0xa02d0c,	/* queue status */
};

enum {						/* scheduler internal data */
	Scdctxdata	= 0x600,
	Scdtxsttsbmap	= 0x7b1,
	Scdtranstbl	= 0x7e0,
};

enum {						/* Icr values */
	Icrfhrx		= 1<<31,		/* Rx DMA (Fhicr[17:16] bits set) */
	Icrhwerr		= 1<<29,		/* DMA hw error */
	Icrucload		= 1<<28,		/* Ucode download */
	Icrfhtx		= 1<<27,		/* Tx DMA (Fhicr[1:0] bits set) */
	Icrscd		= 1<<26,		/* Txq pointer advanced */
	Icrswerr		= 1<<25,		/* Ucode software error */
	Icrfkill		= 1<<7,		/* hw RFKILL switch toggled */
	Icrtempkill 	= 1<<6,		/* high temperature kill */
	Icrwake		= 1<<1,		/* controller waking up */
	Icrucalive		= 1<<0,		/* ucode interrupts once initializes */
};

enum {						/* Fhicr values */
	Fhicrerr		= 1<<31,		/* error */
	Fhicrprio		= 1<<30,		/* high-priority Rx, bypass coalescing */
	Fhicr1		= 1<<17,		/* Rx channel 1 */
	Fhicr0		= 1<<16,		/* Rx channel 0 */
	Fhicrt1		= 1<<1,		/* Tx channel 1 */
	Fhicrt0		= 1<<0,		/* Tx channel 0 */
};

enum {						/* Reset values */
	Resetnevo		= 0x001,
	Resetfnmi		= 0x002,		/* force NMI */
	Resetsw		= 0x080,
	Resetdisabled	= 0x100,		/* master disabled */
	Resetstop		= 0x200,		/* stop master */
};

enum {						/* Config values */
	Configmacsi	= 0x00000100,
	Configradsi	= 0x00000200,
	Confighapwake= 0x00080000,	/* HAP INTA move device L1a->L0s */
	Configeesem	= 0x00200000,	/* EEPROM semaphore */
};

enum {
	Gpctlclkready	= 0x00000001,	/* MAC clock ready */
	Gpctlinitdone	= 0x00000004,	/* Initialization done */
	Gpctlaccreq	= 0x00000008,	/* MAC access request */
	Gpctlsleep	= 0x00000010,	/* going to sleep */
	Gpctlrfkill		= 0x08000000,
};

enum {						/* EEPROM values */
	EEsize		= 2048*2,
	EEradio		= 2*0x48,
};

enum {
	Rqbufsize		= 8192,
	Rqsize		= 256,
};

enum {						/* Ucode values */
	UCinststart	= 0x00000,
	UCinstend		= 0x20000,
	UCinstsize	= UCinstend - UCinststart,
	UCdatastart	= 0x800000,
	UCdataend	= 0x80C000,
	UCdatasize	= UCdataend - UCdatastart,
	UCmaxfilesize	= UCinstsize*2 + UCdatasize*2,
};

typedef struct Ctlr Ctlr;
typedef struct Tq Tq;
typedef struct Tf Tf;
typedef struct Rq Rq;

struct Tq {
	QLock;
	uint		tid;		/* queue id */
	uchar*	bd;		/* base address of the descriptor ring, aligned */
	int		nd;		/* number of descriptors in the ring */
	ushort*	cntbl;	/* count table */
	int		rp;		/* index of the first descriptor used */
	int		wp;		/* index of the last descriptor used + 1 */
	QLock	lock;
	Rendez	rendez;
	Tf*		frames;	/* array with nd elements */
	uchar*	bdna;	/* base address of the descriptor ring, unaligned */
};

struct Tf {
	int		nb;		/* number of blocks in the frame */
	Block*	blocks;	/* list of the blocks */
	int		sp;		/* offset in the first block where the frame starts */
	int		ep;		/* offset in the last block where the frame ends */
	int		len;		/* total length of the frame */
	int		datalen;	/* length of the data part of the frame (ugly) */
	int		clen;		/* current length, used when building the frame */
	ushort	seq;		/* original sequence number */
	int		combine;	/* whether to combine all blocks before sending the frame */
};

struct Rq {
	QLock;
	uint*		bd;		/* base address of the descriptor ring, aligned */
	int		nd;		/* number of descriptors in the ring */
	int		bsize;	/* size of the receive buffers */
	int		rp;		/* index of the first buffer ready to receive data */
	int		wp;		/* index of the last buffer ready to receive data + 1 */
	uint*		rsp;		/* pointer to the status buffer, first int contains hrp */
	Block**	blocks;	/* array with nd elements, nil if the descriptor not ready */
	Rendez	rendez;
	uchar*	bdna;	/* base address of the descriptor ring, unaligned */
	Block*	b;		/* block to read from */
};

struct Ctlr {
	Lock;
	int		status;
	ulong	inuse;
	int		port;
	Pcidev*	pcidev;
	int		irq;
	int*		mem;
	int		rev;
	int		warev;
	ushort	eeprom[EEsize];

	int		ucinstsize;
	uchar	ucinst[UCinstsize];
	int		ucdatasize;
	uchar	ucdata[UCdatasize];
	Rendez	ucliverend;
	Rendez	ucdonerend;

	/* Tx queues */
	Tq		tqs[20];
	uchar*	tqcntbl;	/* 16 tables * 1024 bytes each, aligned */
	uchar*	tqcntblna; /* unaligned */
	uchar*	kw;		/* keep-warm buffer */
	uchar*	kwna;	/* keep-warm buffer, unaligned */

	/* Rx queue */
	Rq		rq;	
};

enum {
	Qdir = 0,
	Qfirstexcl,		/* first with exclusive open */
	Qctl = Qfirstexcl,
	Qrx,
	Qtx0,
	Qtx1,
	Qtx2,
	Qtx3,
	Qtx4,
	Qtx5,
	Qtx6,
	Qtx7,
	Qtx8,
	Qlastexcl,			/* last with exclusive open */
	Qstatus,
	Qucinst,
	Qucdata,
	Qeeprom,
};

static int q2txfifo[] = {
	3, 2, 1, 0, 7, 5, 6,
};

static Dirtab iwldir[] = {
	".", {Qdir, 0, QTDIR}, 0, DMDIR|0555,
	"ctl", {Qctl}, 0, 0666,
	"status", {Qstatus}, 0, 0444,
	"ucinst", {Qucinst}, 0, 0666,
	"ucdata", {Qucdata}, 0, 0666,
	"eeprom", {Qeeprom}, EEsize, 0444,
	"rx", {Qrx}, 0, 0444,
	"tx0", {Qtx0}, 0, 0222,
	"tx1", {Qtx1}, 0, 0222,
	"tx2", {Qtx2}, 0, 0222,
	"tx3", {Qtx3}, 0, 0222,
	"tx4", {Qtx4}, 0, 0222,
	"tx5", {Qtx5}, 0, 0222,
	"tx6", {Qtx6}, 0, 0222,
	"tx7", {Qtx7}, 0, 0222,
	"tx8", {Qtx8}, 0, 0222,
};

enum {				/* ctl commands */
	CMreset,			/* reset the controller */
};

enum {				/* status values */
	Sreset	= 1<<0,	/* reset performed */
	Skproc	= 1<<1,	/* kproc started */
	Sucdone	= 1<<2,	/* transfer of a ucode section is done */
	Sucalive	= 1<<3,	/* the alive notification was received */
};

static Cmdtab ctlmsg[] = {
	CMreset, "reset", 0,
};

static char Ereset[] = "not reset";
static Ctlr *ctlr;
static int debug = 1;

#define dprint if(debug)print
/*
#define csr32r(c, r)	(*((c)->mem+((r)/4)))
#define csr32w(c, r, v)	(*((c)->mem+((r)/4)) = (v))
#define csr32bc(c, r, m) (csr32w((c), (r), (csr32r((c), (r)) & (~m))))
#define csr32bs(c, r, m) (csr32w((c), (r), (csr32r((c), (r)) | (m))))
#define prph32bc(c, r, m) (prph32w((c), (r), (prph32r((c), (r)) & (~m))))
#define prph32bs(c, r, m) (prph32w((c), (r), (prph32r((c), (r)) | (m))))
#define ee16r(c, r) ((c)->eeprom[(r)])
*/

static void iwlinterrupt(Ureg *, void *arg);
static int apminit(Ctlr *ctlr);
static int txinit(Ctlr *ctlr);
static int txstop(Ctlr *ctlr);
static int txreset(Ctlr *ctlr);
static int tqstop(Ctlr *ctlr, int tid);
static int tqreset(Ctlr *ctlr, int tid);
static int tqbwrite(Ctlr *ctlr, int tid, Block *b);
static int rqinit(Ctlr *ctlr);
static int rqstop(Ctlr *ctlr);
static int rqreset(Ctlr *ctlr);
static int rqreplenish(Rq *rq);
static Block *rqbread(Ctlr *ctlr);
static void iwlirqdisable(Ctlr *ctlr);

uint
csr32r(Ctlr *c, uint r)
{
	uint n;

	n = (*((c)->mem+((r)/4)));
	dprint("csr32r(0x%ux) = 0x%ux\n", r, n);
	return n;
}

void
csr32w(Ctlr *c, uint r, uint v)
{
	(*((c)->mem+((r)/4)) = (v));
	dprint("csr32w(0x%ux, 0x%ux)\n", r, v);
}

void
csr32bc(Ctlr *c, uint r, uint m)
{
	uint n;

	n = (*((c)->mem+((r)/4)));
	*((c)->mem+((r)/4)) = n&(~m);
	dprint("csr32bc(0x%ux, 0x%x) = %ux\n", r, m, n&(~m));
}

void
csr32bs(Ctlr *c, uint r, uint m)
{
	uint n;

	n = (*((c)->mem+((r)/4)));
	*((c)->mem+((r)/4)) = n|m;
	dprint("csr32bs(0x%ux, 0x%ux) = %ux\n", r, m, n|m);
}

ushort
ee16r(Ctlr *c, uint r)
{
	uint n;

	n = c->eeprom[r];
	dprint("ee16r(0x%ux) = 0x%ux\n", r, n);
	return n;
}

static int
csr32poll(Ctlr *ctlr, int r, int v, int m, int tout)
{
	int i, d;
	uint n;

	i = 0;
	d = debug; debug = 0;
	do {
		n = csr32r(ctlr, r);
		if ((n & m) == (v & m)) {
			debug = d;
			dprint("csr32poll(0x%ux, 0x%ux, 0x%ux) = %d iteration\n", r, v, m, i);
			return 1;
		}

		microdelay(10);
		i += 10;
	} while (i < tout);

	debug = d;
	dprint("csr32poll(0x%ux, 0x%ux, 0x%ux) = timeout (0x%ux)\n", r, v, m, n);
	return 0;
}

static uint
prph32r(Ctlr *ctlr, int r)
{
	int d;
	uint n;

	d = debug; debug = 0;
	csr32w(ctlr, Prphraddr, r | (3 << 24));	/* 3+1 bytes to read */
	n = csr32r(ctlr, Prphrdata);
	debug = d;
	dprint("prph32r(0x%ux)=0x%ux\n", r, n);
	return n;
}

static void
prph32w(Ctlr *ctlr, int r, int v)
{
	int d;

	d = debug; debug = 0;
	csr32w(ctlr, Prphwaddr, r | (3<<24));
	csr32w(ctlr, Prphwdata, v);
	debug = d;
	dprint("prph32w(0x%ux, 0x%ux)\n", r, v);
}

void
prph32bc(Ctlr *c, uint r, uint m) 
{
	int d;
	uint n;

	d = debug; debug = 0;
	n = prph32r(c, r);
	prph32w(c, r, n&(~m));
	debug = d;
	dprint("prph32bc(0x%ux, 0x%ux) = 0x%ux\n", r, m, n&(~m));
}

void
prph32bs(Ctlr *c, uint r, uint m)
{
	int d;
	uint n;

	d = debug; debug = 0;
	n = prph32r(c, r);
	prph32w(c, r, n|m);
	debug = d;
	dprint("prph32bs(0x%ux, 0x%ux) = 0x%ux\n", r, m, n|m);
}

static uint
sram32r(Ctlr *ctlr, uint r)
{
	uint n;
	int d;

	d = debug; debug = 0;
	csr32w(ctlr, Hbusraddr, r);
	n = csr32r(ctlr, Hbusrdata);
	debug = d;
	dprint("sram32r(%ux) = %ux\n", r, n);
	return n;
}

static void
sram32w(Ctlr *ctlr, uint r, uint v)
{
	int d;

	d = debug; debug = 0;
	csr32w(ctlr, Hbuswaddr, r);
	csr32w(ctlr, Hbuswdata, v);
	debug = d;
	dprint("sram32r(%ux, %ux)\n", r, v);
}


static int
nicgrab(Ctlr *ctlr)
{
	int n;

	dprint("-- nicgrab\n");
	csr32bs(ctlr, Gpctl, Gpctlaccreq);
	n = csr32poll(ctlr, Gpctl, Gpctlclkready, Gpctlclkready|Gpctlsleep, 50);
	if (!n)
		dprint("can't grab nic\n");

	return n;
}

static void
nicrelease(Ctlr *ctlr)
{
	csr32bc(ctlr, Gpctl, Gpctlaccreq);
	dprint("-- nicrelease\n");
}

static int
eeload(Ctlr *ctlr)
{
	int i, n, r;
	int d;

	dprint("-- eeload start\n");
	/* verify EEPROM signature */
	if ((csr32r(ctlr, EEgp) & 6) == 0)
		return 0;

	/* acquire EEPROM semaphore */
	for(i = 0; i < 1000; i++) {
		csr32bs(ctlr, Config, Configeesem);
		if (csr32poll(ctlr, Config, Configeesem, Configeesem, 10000))
			break;
	}
	if (i >= 1000)
		return 0;

	d = debug; debug = 0;
	/* read the EEPROM content */
	for(n = 0; n < EEsize; n++) {
		csr32w(ctlr, EEreg, n << 2);
		csr32bc(ctlr, EEreg, 2);		/* command */
		for(r=0, i=0; i < 500; i++) {
			r = csr32r(ctlr, EEreg);
			if (r  & 1)
				break;

			microdelay(10);
		}

		if (!(r&1)) {
			debug = d;
			return 0;
		}

		ctlr->eeprom[n] = r >> 16;		
	}
	debug = d;

	/* release semaphore */
	csr32bc(ctlr, Config, Configeesem);
	dprint("-- eeload end\n");
	return 1;
}

static int
recvucalive(void *a)
{
	int rp;
	Block *b;
	Ctlr *ctlr;
	Rq *rq;

	ctlr = a;
//	lock(ctlr);
	rq = &ctlr->rq;
	for(rp = rq->rp; rp != ((*rq->rsp) & 0xFFF);) {
		b = rq->blocks[rp];
		if (b->rp[5] == 0x01) {
//			unlock(ctlr);
			return 1;
		}

		rp++;
		if (rp >= rq->nd)
			rp = 0;
	}

//	unlock(ctlr);
	return 0;
}

static void
iwlrproc(void* arg)
{
	int i, n, rp, found;
	uint scdbaddr, p;
	Block *b;
	Ctlr *ctlr;
	Rq *rq;

	dprint("iwlrpoc start\n");
	ctlr = arg;
	for(;;) {
		sleep(&ctlr->ucliverend, recvucalive, ctlr);
		lock(ctlr);
		rq = &ctlr->rq;
		coherence();
		dprint("got ucode alive notification rp %d hrp %d\n", rq->rp, (*rq->rsp) & 0xFFF);
		for(found = 0, rp = rq->rp; rp != ((*rq->rsp) & 0xFFF);) {
			b = rq->blocks[rp];
			if (b && b->rp[4] == 0x01) {
				found = 1;
				break;
			}

			rp++;
			if (rp >= rq->nd)
				rp = 0;
		}

		if (!found) {
			unlock(ctlr);
			continue;
		}

		if (!nicgrab(ctlr)) {
			print("can't grab nic\n");
			unlock(ctlr);
			continue;
		}

		scdbaddr = prph32r(ctlr, Scdsaddr);
		for(p = Scdctxdata; p < Scdtxsttsbmap; p += 4)
			sram32w(ctlr, p + scdbaddr, 0);

		for(; p < Scdtranstbl; p += 4)
			sram32w(ctlr, p + scdbaddr, 0);

/* most likely a bug in the linux driver, but this is missing 
		for(; p < Scdtranstbl + 40; p += 4)
			sram32w(ctlr, p + scdbaddr, 0);
*/

		prph32w(ctlr, Scddaddr, (PCIWADDR(ctlr->tqcntbl))>>10);
		prph32w(ctlr, Scdqchain, 0xFFFFF);
		prph32w(ctlr, Scdaggr, 0);
		/* initiate the queues */
		for(i = 0; i < 20; i++) {
			prph32w(ctlr, Scdqhrp + 4*i, 0);
			csr32w(ctlr, Tqhwp, 0 | (i << 8));
			sram32w(ctlr, scdbaddr + Scdctxdata + i*8, 0);
			sram32w(ctlr, scdbaddr + Scdctxdata + i*8 + 4, 64 |	/* win size */
				(64 << 16));							/* frame limit */
		}
		prph32w(ctlr, Scdintrmask, 0x1FFFFF);
		prph32w(ctlr, Tqfact, 0xFF);
		csr32w(ctlr, Tqhwp, 0 | (4<<8));		/* cmd queue */
		prph32w(ctlr, Scdqhrp + 4*4, 0);
	
		/* map qos queues to fifos one-to-one */
		for(i = 0; i < 7; i++) {
			n = q2txfifo[i];
			prph32w(ctlr, Scdqstatus + i*4, (1<< 3) |	/* active */
				(n << 0) |					/* tx fifo to use */
				(1 << 4) |					/* window size left */
				0xFF0000);				/* stts reg mask */
		}

		nicrelease(ctlr);
		ctlr->status |= Sucalive;
		unlock(ctlr);
		wakeup(&ctlr->rq.rendez);
	}
}

static void
iwlreset(void)
{
	Pcidev *p;
	void *mem;
	static int isreset = 0;

	if (isreset)
		return;

	if (getconf("*noiwlprobe"))
		return;

	isreset = 1;
	p = nil;
	while (p = pcimatch(p, 0, 0)) {
		if (p->did != 0x4237 || p->vid != 0x8086)
			continue;

		mem = vmap(p->mem[0].bar & ~0x0F, p->mem[0].size);
		if (mem == nil) {
			print("iwl: can't map %8.8luX\n", p->mem[0].bar);
			continue;
		}

		ctlr = malloc(sizeof(Ctlr));
		ctlr->port = p->mem[0].bar & ~0x0F;
		ctlr->pcidev = p;
		ctlr->mem = mem;
		ctlr->irq = p->intl;

		pcisetbme(p);
		/* TODO: set_dma_mask? */
		iwlirqdisable(ctlr);
		ctlr->rev = csr32r(ctlr, Revision);
		ctlr->warev = csr32r(ctlr, Warevision);
		if (!apminit(ctlr))
			goto error;

		if (!eeload(ctlr))
			goto error;

		if (!rqinit(ctlr))
			goto error;

		if (!txinit(ctlr))
			goto error;

		print("iwl5000 rev %d warev %d irq %d\n", ctlr->rev, ctlr->warev, ctlr->irq);
		intrenable(ctlr->irq, iwlinterrupt, ctlr, p->tbdf, "iwl");
	}

	return;

error:
	vunmap(mem, p->mem[0].size);
	free(ctlr);
}

void
iwlinit(void)
{
}

Chan *
iwlattach(char *spec)
{
	if (ctlr == nil)
		error(Enodev);

	lock(ctlr);
	if (!(ctlr->status & Skproc)) {
		kproc("#wrproc", iwlrproc, ctlr);
		ctlr->status |= Skproc;
	}
	unlock(ctlr);
//	rqreplenish(&ctlr->rq);

	return devattach('w', spec);
}

static Walkqid*
iwlwalk(Chan *c, Chan *nc, char **name, int nname)
{
	return devwalk(c, nc, name, nname, iwldir, nelem(iwldir), devgen);
}

static int
iwlstat(Chan *c, uchar *db, int n)
{
	return devstat(c, db, n, iwldir, nelem(iwldir), devgen);
}

Chan*
iwlopen(Chan *c, int omode)
{
	ulong bit, inuse, path;

	path = c->qid.path;
	if (path < Qlastexcl) {
		bit = 1<<(path - Qfirstexcl);
		lock(ctlr);
		inuse = ctlr->inuse & bit;
		ctlr->inuse |= bit;
		unlock(ctlr);

		if (inuse)
			error(Einuse);
	}

	c = devopen(c, omode, iwldir, nelem(iwldir), devgen);
	return c;
}

void
iwlclose(Chan *c)
{
	ulong bit, path;

	path = c->qid.path;
	if (path < Qlastexcl) {
		bit = 1<<(path - Qfirstexcl);
		lock(ctlr);
		ctlr->inuse &= ~bit;
		unlock(ctlr);
	}
}

long
iwlread(Chan *c, void *a, long n, vlong offset)
{
	switch ((int) c->qid.path) {
	default:
		error(Eio);
		break;

	case Qdir:
		n = devdirread(c, a, n, iwldir, nelem(iwldir), devgen);
		break;

	case Qctl:
		break;

	case Qstatus:
		break;

	case Qucinst:
		if (n+offset > ctlr->ucinstsize)
			n = ctlr->ucinstsize - offset;
		if (n > 0)
			memmove(a, ctlr->ucinst+offset, n);
		else
			n = 0;
		break;

	case Qucdata:
		if (n+offset > ctlr->ucdatasize)
			n = ctlr->ucdatasize - offset;
		if (n > 0)
			memmove(a, ctlr->ucdata+offset, n);
		else
			n = 0;
		break;

	case Qeeprom:
		if (offset < sizeof(ctlr->eeprom)) {
			if (n > sizeof(ctlr->eeprom) -offset)
				n = sizeof(ctlr->eeprom) - offset;
			memmove(a, ((uchar *) ctlr->eeprom) + offset, n);
		} else
			n = 0;
		break;

	case Qrx:
		/* FIXME: racy, but you better read from one proc */
		if (!ctlr->rq.b)
			ctlr->rq.b = rqbread(ctlr);

		if (n > BLEN(ctlr->rq.b))
			n = BLEN(ctlr->rq.b);

		memmove(a, ctlr->rq.b->rp, n);
		ctlr->rq.b->rp += n;
		if (BLEN(ctlr->rq.b) == 0) {
			freeb(ctlr->rq.b);
			ctlr->rq.b = nil;
		}
		break;
	}

	return n;
}

static Block*
iwlbread(Chan* c, long n, ulong offset)
{
	dprint("iwlbread: path %lld\n", c->qid.path);
	if ((int) c->qid.path == Qrx) {
		dprint("iwlbread: rx\n");
		return rqbread(ctlr);
	} else
		return devbread(c, n, offset);
}

static void
iwlirqenable(Ctlr *ctlr)
{
	dprint("-- iwlirqenable start\n");
//	csr32w(ctlr, Intmask, Icrfhrx | Icrhwerr | Icrucload | Icrfhtx | Icrscd |
//		Icrfkill | Icrtempkill | Icrwake | Icrucalive);
	csr32w(ctlr, Intmask, 0xbe0000cb);
	dprint("-- iwlirqenable end\n");
}

static void
iwlirqdisable(Ctlr *ctlr)
{
	dprint("-- iwlirqdisable start\n");
	csr32w(ctlr, Intmask, 0);
	csr32w(ctlr, Icr, 0xFFFFFFFF);		/* ack/reset all flags */
	csr32w(ctlr, Fhicr, 0xFFFFFFFF);
	dprint("-- iwlirqdisable end\n");
}

static int
ucdone(void *a)
{
	Ctlr *ctlr;

	ctlr = a;
	return (ctlr->status & Sucdone) != 0;
}

static int
ucloadsection(Ctlr *ctlr, uchar *src, int srclen, ulong dst)
{
	dprint("ucloadsection start src %lx slen %d dst %lx\n", PCIWADDR(src), srclen, dst);
	lock(ctlr);
	ctlr->status &= ~Sucdone;
	if (!nicgrab(ctlr)) {
		dprint("ucloadsection end with error\n");
		unlock(ctlr);
		return 0;
	}

	csr32w(ctlr, Tsvcconfig, 0);
	csr32w(ctlr, Tsvcsaddr, dst);
	csr32w(ctlr, Tsvcf0, PCIWADDR(src));
	csr32w(ctlr, Tsvcf1, srclen);
	csr32w(ctlr, Tsvcsts, (1<<20) | (1<<12) | 3);
	csr32w(ctlr, Tsvcconfig, 0x80100000);
	nicrelease(ctlr);
	unlock(ctlr);
	tsleep(&ctlr->ucdonerend, ucdone, ctlr, 5000);

	dprint("ucloadsection end\n");
	return ucdone(ctlr);
}

static int
ucload(Ctlr *ctlr)
{
	coherence();
	/* load the instructions */
	dprint("load instructions...\n");
	if (!ucloadsection(ctlr, ctlr->ucinst, ctlr->ucinstsize, 0))
		return 0;

	/* load the data */
	dprint("load data...\n");
	if (!ucloadsection(ctlr, ctlr->ucdata, ctlr->ucdatasize, 0x800000))
		return 0;

	return 1;
}

static int
apminit(Ctlr *ctlr)
{
	dprint("-- apminit start\n");
	csr32bs(ctlr, Giobits, 0x20000000);
	csr32bs(ctlr, Giobits, 0x00800000);		/* disable L0s without affecting L1 */
	csr32bs(ctlr, Hpetmem, 0xFFFF0000);	/* set FH wait to max */
	csr32bs(ctlr, Config, Confighapwake);	/* enable HAP INTA L1a->L0s */
	csr32bs(ctlr, Anlgpll, 0x00880300);
	csr32bs(ctlr, Gpctl, Gpctlinitdone);		/* init done, D0U->D0A */

	/* wait for clock stabilization */
	if (!csr32poll(ctlr, Gpctl, Gpctlclkready, Gpctlclkready, 25000))
		goto error;

	if (!nicgrab(ctlr))
		goto error;

	prph32w(ctlr, Apmgclkenable, 0x200);	/* enable DMA */
	microdelay(20);
	prph32bs(ctlr, Apmgpcistt, 0x800);		/* disable L1-Active */
	nicrelease(ctlr);
	dprint("-- apminit end\n");
	return 1;

error:
	dprint("-- apminit end with error\n");
	return 0;
}

static int
apmreset(Ctlr *ctlr)
{
	dprint("-- apmreset start\n");
	csr32bs(ctlr, Reset, Resetstop);		/* set stop master bit */
	if (!csr32poll(ctlr, Reset, Resetdisabled, Resetdisabled, 100))
		goto error;

	csr32bs(ctlr, Reset, Resetsw);
	microdelay(10);
	csr32bs(ctlr, Anlgpll, 0x00880300);
	csr32bs(ctlr, Gpctl, Gpctlinitdone);		/* init done, D0U->D0A */

	/* wait for clock stabilization */
	if (!csr32poll(ctlr, Gpctl, Gpctlclkready, Gpctlclkready, 25000)) {
		dprint("apmreset: no clock stabilization\n");
		goto error;
	}

	if (!nicgrab(ctlr))
		goto error;

	prph32w(ctlr, Apmgclkenable, 0x200);	/* enable DMA */
	microdelay(20);
	prph32bs(ctlr, Apmgpcistt, 0x800);		/* disable L1-Active */
	nicrelease(ctlr);
	dprint("-- apmreset end\n");
	return 1;

error:
	dprint("-- apmreset end with error\n");
	return 0;
}

#ifdef notused
static int
apmstop(Ctlr *ctlr)
{
	lock(ctlr);
	csr32bs(ctlr, Reset, Resetstop);		/* set stop master bit */
	if (!csr32poll(ctlr, Reset, Resetdisabled, Resetdisabled, 100))
		goto error;

	csr32bs(ctlr, Reset, Resetsw);
	microdelay(10);
	csr32bc(ctlr, Gpctl, Gpctlinitdone);
	unlock(ctlr);
	return 1;

error:
	unlock(ctlr);
	return 0;
}
#endif

static int
iwlrestart(Ctlr *ctlr)
{
	ushort radio;

	dprint("iwlrestart: start\n");
	rqreplenish(&ctlr->rq);
	lock(ctlr);
	ctlr->status &= ~Sucalive;
	csr32w(ctlr, Reset, Resetnevo);
	iwlirqdisable(ctlr);
	csr32bc(ctlr, Gpctl, 8);
	if (!nicgrab(ctlr))
		goto error;

	txstop(ctlr);
	if (!rqstop(ctlr))
		goto error;

	prph32w(ctlr,  Apmgclkdisable, 0x200);
	nicrelease(ctlr);
	microdelay(5);

	if (!apmreset(ctlr))
		goto error;

	if (!csr32r(ctlr, Gpctl) & Gpctlrfkill) {
		dprint("rfkill %x\n", csr32r(ctlr, Gpctl));
		goto error;
	}

	csr32w(ctlr, Icr, 0xFFFFFFFF);
	if (!apminit(ctlr))
		goto error;

	csr32w(ctlr, Intcoalescing, 512/32);
	if (!nicgrab(ctlr))
		goto error;

	prph32w(ctlr, Apmgpwr, prph32r(ctlr, Apmgpwr) & ~0x03000000);
	if (pcicfgr16(ctlr->pcidev, 0xF0) & 0x02)
		csr32bs(ctlr, Gio, 2);		/* L0s disabled, L1A enabled */
	else
		csr32bc(ctlr, Gio, 2);		/* L0s enabled, L1a disabled */

#ifdef right
	radio = ee16r(ctlr, EEradio);
	if (radio&3 <= 3)
		csr32bs(ctlr, Config, radio & 0x1F);
#else
	csr32bs(ctlr, Config, 3);
#endif

	csr32bs(ctlr, Config, Configradsi | Configmacsi);
	prph32bs(ctlr, Apmgpwr, 0x00400000);
	if (!rqreset(ctlr))
		goto relerror;

	csr32w(ctlr, Intcoalescing, 0x40);	// do we need this???
	if (!txreset(ctlr))
		goto relerror;

	nicrelease(ctlr);
	unlock(ctlr);
	coherence();

	csr32w(ctlr, Gp1clr, 2);
	csr32w(ctlr, Gp1clr, 4);
	csr32w(ctlr, Icr, 0xFFFFFFFF);
	iwlirqenable(ctlr);
	csr32w(ctlr, Gp1clr, 2);
	csr32w(ctlr, Gp1clr, 2);
	if (!ucload(ctlr))
		return 0;

	csr32w(ctlr, Reset, 0);

	return 1;

relerror:
	nicrelease(ctlr);

error:
	dprint("iwlrestart: error\n");
	unlock(ctlr);
	return 0;
}

static void
iwlinterrupt(Ureg *, void *arg)
{
	int d;
	uint icr, fhicr;
	Ctlr *ctlr;

	ctlr = arg;
	ilock(ctlr);
	d = debug;
	debug = 0;
	icr = csr32r(ctlr, Icr);
	fhicr = csr32r(ctlr, Fhicr);
	if (!icr && !fhicr) {
		debug = d;
		iunlock(ctlr);
		return;
	}

	csr32w(ctlr, Icr, icr);				/* ack/reset the received interrupts */
	csr32w(ctlr, Fhicr, fhicr);			/* ack/reset the received Fh interrupts */
//	dprint("iwlinterrupt: icr %ux fhicr %ux\n", icr, fhicr);

	/* some of the bits are both icr and fhicr and because the ack is not atomic,
	     they can be different. make sure they are in sync */
	if (fhicr & (Fhicr0 | Fhicr1 | Fhicrprio))
		icr |= Icrfhrx;

	if (fhicr & (Fhicrt0 | Fhicrt1))
		icr |= Icrfhtx;

	/* now handle the events */
	if (icr & (Icrhwerr | Icrswerr | Icrtempkill | Icrfkill)) {
		debug = d;
		print("iwl: error %x\n", icr);
		iwlirqdisable(ctlr);
		iunlock(ctlr);
		return;
	}

	if (icr & Icrwake)
		print("iwl: wakeup detected, not handled\n");

//	if (icr & Icrucalive) {
//		dprint("iwlinterrupt: ucode alive\n");
//	}

	if (icr & Icrfhrx) {
		if (!(ctlr->status & Sucalive))
			wakeup(&ctlr->ucliverend);
		else
			wakeup(&ctlr->rq.rendez);
	}

	if (icr & Icrfhtx) {
		ctlr->status |= Sucdone;
		wakeup(&ctlr->ucdonerend);
	}

//	iwlirqenable(ctlr);
	debug = d;
	iunlock(ctlr);
}

static long
txwrite(Ctlr *ctlr, int qno, void *data, long datalen, vlong offset)
{
	USED(qno);
	USED(data);
	USED(datalen);
	USED(offset);

	if (!(ctlr->status & Sreset))
		error(Ereset);

	return 0;
}

long
iwlwrite(Chan *c, void *a, long n, vlong offset)
{
	Cmdtab *ct;
	Cmdbuf *cb;
	Block *b;

	switch ((int) c->qid.path) {
	default:
		error(Eio);
		break;

	case Qucinst:
		if (offset+n>sizeof(ctlr->ucinst))
			error("writing outside ucinst boundaries");
		memmove(ctlr->ucinst + offset, a, n);
		if (offset+n > ctlr->ucinstsize)
			ctlr->ucinstsize = offset+n;
		return n;

	case Qucdata:
		if (offset+n>sizeof(ctlr->ucdata))
			error("writing outside ucdata boundaries");
		memmove(ctlr->ucdata + offset, a, n);
		if (offset+n > ctlr->ucdatasize)
			ctlr->ucdatasize = offset+n;
		return n;

	case Qctl:
		cb = parsecmd(a, n);
		if (waserror()) {
			free(cb);
			nexterror();
		}

		ct = lookupcmd(cb, ctlmsg, nelem(ctlmsg));
		if (ct->index == CMreset)
			iwlrestart(ctlr);

		poperror();
		free(cb);
		return n;

	case Qtx0:
	case Qtx1:
	case Qtx2:
	case Qtx3:
	case Qtx4:
	case Qtx5:
	case Qtx6:
	case Qtx7:
		b = allocb(n);
		memmove(b->rp, a, n);
		b->wp += n;
		return tqbwrite(ctlr, c->qid.path - Qtx0, b);
	}

	return 0;
}

static long
iwlbwrite(Chan* c, Block* bp, ulong offset)
{
	switch ((int) c->qid.path) {
	default:
		return devbwrite(c, bp, offset);

	case Qtx0:
	case Qtx1:
	case Qtx2:
	case Qtx3:
	case Qtx4:
	case Qtx5:
	case Qtx6:
	case Qtx7:
		return tqbwrite(ctlr, c->qid.path - Qtx0, bp);
	}
}

static int
iwlwstat(Chan *c, uchar *db, int n)
{
	Dir *d;

	if (c->qid.path!=Qucinst && c->qid.path!=Qucdata)
		return devwstat(c, db, n);

	d = nil;
	if (waserror()) {
		free(d);
		nexterror();
	}

	n = convM2D(db, n, &d[0], (char *) &d[1]);
	if (n == 0)
		error(Eshortstat);

	if (d->length == ~0)
		error(Eperm);

	if (c->qid.path == Qucinst)
		ctlr->ucinstsize = 0;
	else
		ctlr->ucdatasize = 0;

	poperror();
	free(d);
	return n;
}

static int
tqinit(Tq *tq, int tid, ushort *cntbl)
{
	memset(tq, 0, sizeof(*tq));
	tq->nd = 256;		/* can't be changed */
	tq->tid = tid;
	tq->bdna = malloc(tq->nd * 128 + 256);
	tq->bd = (uchar *) ROUND((ulong) tq->bdna, 256);
	tq->cntbl = cntbl;
	tq->rp = 0;
	tq->wp = 0;
	tq->frames = mallocz(tq->nd * sizeof(Tf), 1);

	return 1;
}

/* should be called with the nic grabbed */
static int
tqreset(Ctlr *ctlr, int tid)
{
	Tq *tq;

	tq = &ctlr->tqs[tid];
	dprint("-- tqreset %d start bd %p/%lx cntbl %p\n", tid, tq->bd, PCIWADDR(tq->bd), tq->cntbl);
	prph32bc(ctlr, Tqfact, 1<<tid);					/* disable DMA channel */
	csr32w(ctlr, Tqconfig + 0x20*tid, 0);			/* disable DMA channel ??? */
	csr32poll(ctlr, Tqstatus, 0x1010000<<tid, 0x1010000<<tid, 200);/* wait until the channel is idle */
	csr32w(ctlr, Tqbd + tid*4, (PCIWADDR(tq->bd)) >> 8);
	csr32w(ctlr, Tqhwp, tid<<8);					/* reset hwp */
	tq->rp = 0;
	tq->wp = 0;
	csr32w(ctlr, Tqconfig + 0x20*tid, 0x80000008);	/* enable DMA */
	dprint("-- tqreset %d end\n", tid);
	return 1;
}

/* should be called with the nic grabbed */
static int
tqstop(Ctlr *ctlr, int tid)
{
	dprint("-- tqstop %d start\n", tid);
	prph32bc(ctlr, Tqfact, 1<<tid);				/* disable DMA channel */
	csr32w(ctlr, Tqconfig + 0x20*tid, 0);		/* disable DMA channel ??? */
	csr32poll(ctlr, Tqstatus, 0x11<<tid, 0x01010000<<tid, 200);/* wait until the channel is idle */
	dprint("-- tqstop %d end\n", tid);
	return 1;
}

static int
tqavail(void *a)
{
	int n;
	Tq *tq;

	tq = a;
	n = tq->wp - tq->rp;
	if (n < 0)
		n += tq->nd;

	dprint("tqavail: rp %d wp %d used %d\n", tq->rp, tq->wp, n);
	return n < tq->nd - 2;
}

/* called with the lock held */
static int
tqsetupframe(Tq *tq)
{
	int n, rp, len;
	ulong addr;
	Tf *f;
	Block *b, *b1, *bn;
	uint *fd;
	ushort *sp;

	f = &tq->frames[tq->wp];
	f->combine |= f->nb > 20;
	dprint("-- tqsetupframe start tq %d frame %d combine %d\n", tq->tid, tq->wp, f->combine);
	if (f->combine) {
		b = allocb(f->len + 4);
		b->rp = (uchar *) ROUND((ulong) b->rp, 4);
		rp = f->sp;
		b1 = f->blocks;
		n = 0;
		while (b1 != nil) {
			if (b1->next)
				len = (b1->wp - b1->rp) - rp;
			else
				len = f->ep - rp;

			memmove(b->rp + n, b1->rp + rp, len);
			n += len;
			rp = 0;
			bn = b1->next;
			freeb(b1);
			b1 = bn;
		}

		b->wp = b->rp + f->len;
		f->blocks = b;
		f->nb = 1;
		f->sp = 0;
		f->ep = b->wp - b->rp;
	}

	/* save and replace the sequence */
	sp = (ushort *) (f->blocks->rp + f->sp + 2);	/* address of the frame sequence */
	f->seq = *sp;
	*sp = (tq->tid<<10) | tq->wp;
	dprint("tqsetupframe: seq %ux\n", f->seq);

	/* setup the buffer descriptors */
	fd = (uint *) (tq->bd + 128*tq->wp + 4);
	for(n = 0, b = f->blocks, rp = f->sp; b != nil; rp = 0, n++, b = b->next) {
		addr = PCIWADDR(b->rp + rp);
		if (b->next)
			len = (b->wp - b->rp) - rp;
		else
			len = f->ep - rp;

		{
			int i;
			for(i = 0; i < len; i++)
				dprint("%02ux%c", b->rp[rp + i], (i&0x0f)==0x0f?'\n':' ');
			dprint("\n");
		}

		if (n&1) {
			fd[1] &= 0xFFFF;
			fd[1] |= (addr & 0xFFFF) << 16;
			fd[2] = (len << 20) | (addr >> 16);
			fd += 3;
		} else {
			fd[0] = addr;
			fd[1] = len << 4;
		}

		dprint("tqsetupframe: bd %p fd %p index %d len %d addr %lux %08ux %08ux %08ux\n",
			tq->bd, fd, n, len, addr, fd[0], fd[1], fd[2]);

	}

	*(uint *) (tq->bd + 128*tq->wp) = (f->nb << 24); /* is that right?  | (16 - f->len&15); */
	dprint("tqsetupframe: wp %d numtbs %ux\n", tq->wp, *(uint *) (tq->bd + 12*tq->wp));

	/* setup the count in the count table */
	tq->cntbl[tq->wp] = f->datalen;
	dprint("tqsetupframe: cntbl %p/%p cnt %d\n", tq->cntbl, &tq->cntbl[tq->wp], tq->cntbl[tq->wp]);
	if (tq->wp < 64) {
		/* for some reason the value of the first 64 entries needs to be repeated */
		tq->cntbl[tq->wp + 64] = f->datalen;
	}
	coherence();

	dprint("-- tqsetupframe end tq %d frame %d\n", tq->tid, tq->wp);
	return 1;
}

static int
tqbwrite(Ctlr *ctlr, int tid, Block *b)
{
	int len;
	uchar *p;
	Block *b1;
	Tq *tq;
	Tf *f;

	dprint("-- tqbwrite start tq %d len %ld\n", tid, b->wp - b->rp);
	tq = &ctlr->tqs[tid];
	p = b->rp;

again:
	sleep(&tq->rendez, tqavail, tq);
	qlock(tq);
	while (p < b->wp) {
		if (!tqavail(tq)) {
			qunlock(tq);
			goto again;
		}

		f = &tq->frames[tq->wp];
		if (f->len <= 0) {
			/* This is the first block of a new frame. The first two bytes contain the
			     frame's length, the next two bytes the length of the data */
			if (b->wp-p < 4) {
				qunlock(tq);
				error("frame len across blocks");
			}

			f->len = p[0] | (p[1]<<8);
			f->datalen = p[2] | (p[3]<<8);
			dprint("    start frame %d len %d datalen %d\n", tq->tid, f->len, f->datalen);
			f->clen = 0;
			p += 4;
			f->sp = p - b->rp;
			f->nb = 1;
			f->blocks = b;
			f->combine = (b->wp - p) < 4;
		} else {
			for(b1 = f->blocks; b1->next != nil; b1 = b1->next)
				;

			b1->next = b;
		}
		b->next = nil;
		_xinc(&b->ref);
		f->combine |= ((int) p & 3) || ((b->wp - p) & 3);	/* combine if the block is not aligned */
		len = b->wp - p;
		if (len > f->len - f->clen)
			len = f->len - f->clen;

		f->ep = (p + len) - b->rp;
		f->clen += len;
		if (f->clen == f->len) {
			tqsetupframe(tq);
			dprint("    end frame %d\n", tq->wp);
			tq->wp++;
			/* do we need to grab the nic??? */
			csr32w(ctlr, Tqhwp, tq->wp | (tq->tid<<8));	/* tell the device the new wp */
		}

		p += len;
		dprint("tqbwrite p %p wp %p len %d\n", p, b->wp, len);
	}

	qunlock(tq);
	dprint("-- tqbwrite end tq %d\n", tid);
	return BLEN(b);		
}

static ushort
tqframefree(Ctlr *ctlr, ushort seq)
{
	int i, n, tid;
	Block *b, *bn;
	Tq *tq;
	Tf *f;

	tid = seq>>10;
	n = seq & 0x3FF;
	if (tid > 16) {
		print("invalid seq: %x", seq);
		return 0xFFFF;
	}

	tq = &ctlr->tqs[tid];
	dprint("-- tqframefree start rp %d tid %d n %d\n", tq->rp, tid, n);
	if (tq->nd <= n) {
		print("invalid seq: %x", seq);
		return 0xFFFF;
	}

	qlock(tq);
	f = &tq->frames[n];
	f->len = 0;
	b = f->blocks;
	while (b != nil) {
		bn = b->next;
		freeb(b);
		b = bn;
	}
	f->blocks = nil;
	for(i = tq->rp; i < n; i++) {
		dprint("	f[%d].len = %d\n", i, tq->frames[i].len);
		if (tq->frames[i].len > 0)
			break;
	}

	dprint("tqframefree: i %d n %d\n", i, n);
	if (i == n)
		tq->rp = n + 1;

	seq = f->seq;
	qunlock(tq);
	dprint("-- tqframefree end seq %d\n", seq);
	return seq;
}

static int
txinit(Ctlr *ctlr)
{
	int i;

	/* allocate the count table */
	ctlr->tqcntblna = malloc((nelem(ctlr->tqs) + 1)*1024);
	ctlr->tqcntbl = (uchar *) ROUND((ulong) ctlr->tqcntblna, 1024);
	dprint("txinit: tqcntbl %p\n", ctlr->tqcntbl);
	/* initialize the Tx queues */
	for(i = 0; i < nelem(ctlr->tqs); i++)
		tqinit(&ctlr->tqs[i], i, (ushort *) &ctlr->tqcntbl[1024 * i]);

	/* allocate the keep-warm buffer */
	ctlr->kwna = malloc(8192);
	ctlr->kw = (uchar *) ROUND((ulong) ctlr->kwna, 4096);

	return 1;
}

static int
txreset(Ctlr *ctlr)
{
	int i;

	dprint("-- txreset start\n");
	prph32w(ctlr, Tqfact, 0);					/* disable all DMA channels */
	/* reset the Tx queues */
	for(i = 0; i < nelem(ctlr->tqs); i++)
		tqreset(ctlr, i);

	/* reset the keep-warm buffer */
	csr32w(ctlr, Kwaddr, (PCIWADDR(ctlr->kw))>>4);
	dprint("-- txreset end\n");
	return 1;
}

static int
txstop(Ctlr *ctlr)
{
	int i;

	dprint("-- txstop start\n");
	for(i = 0; i < nelem(ctlr->tqs); i++)
		if (!tqstop(ctlr, i))
			break;

	dprint("-- txstop end\n");
	return i >= nelem(ctlr->tqs);
}

/* should be called with the lock held */
static int
rqalloc(Rq *rq, int n)
{
	Block *b;

	if (rq->blocks[n] != nil)
		print("rqalloc: bug %d rp %d wp %d rsp %d\n", n, rq->rp, rq->wp, *rq->rsp);

	b = allocb(rq->bsize + 256);
	b->rp = (uchar *) ROUND((ulong) b->rp, 256);
	b->wp = b->rp;
	rq->blocks[n] = b;
	rq->bd[n] = (PCIWADDR(b->rp)) >> 8;

	return 1;
}

/* should be called with the lock held */
static int
rqreplenish(Rq *rq)
{
	int n, wp;

	dprint("-- rqreplenish start rp %d wp %d\n", rq->rp, rq->wp);
	n = rq->wp - rq->rp;
	if (n < 0)
		n += rq->nd;

	if (n > rq->nd/2)
		return 0;

	wp = rq->rp - 1;
	if (wp < 0)
		wp = rq->nd - 1;

	wp &= ~7;
	n = rq->wp;
	while (n != wp) {
		rqalloc(rq, n);
		n++;
		if (n >= rq->nd)
			n = 0;
	}

	n = rq->wp != wp;
	rq->wp = wp;
	coherence();
	dprint("-- rqreplenish end rp %d wp %d\n", rq->rp, rq->wp);	
	return n;
}

static int
rqinit(Ctlr *ctlr)
{
	Rq *rq;

	rq = &ctlr->rq;
	memset(rq, 0, sizeof(*rq));
	rq->nd = 256;		/* if changed, rqreset needs change */
	rq->bsize = 8192;	/* if changed, rqresets needs change */
	rq->bdna = malloc(rq->nd*4 + 264);
	rq->bd = (uint *) ROUND((ulong) rq->bdna, 256);
	rq->rsp = rq->bd + rq->nd;
	rq->blocks = malloc(rq->nd * sizeof(Block *));
	return 1;
}

/* should be called with the nic grabbed */
static int
rqreset(Ctlr *ctlr)
{
	int i;
	Rq *rq;

	rq = &ctlr->rq;
	dprint("-- rqreset start bd %p/%lx st %lx\n", rq->bd, PCIWADDR(rq->bd), PCIWADDR(rq->rsp));
	csr32w(ctlr, Rqconfig, 0);					/* disable DMA */
	if (!csr32poll(ctlr, Rqstatus, 1<<24, 1<<24, 1000)) {	/* wait until DMA stopped */
		print("can't stop Rx DMA: %x\n", csr32r(ctlr, Rqstatus));
		return 0;
	}

	/* replenish the buffer descriptor from 0 to rq->rp */
	for(i = 0; i < rq->rp; i++)
		rqalloc(rq, i);

	rq->rp = 0;
	*rq->rsp = 0;
	csr32w(ctlr, Rqhwp, 0);				/* reset hwp */
	csr32w(ctlr, Rqbd, (PCIWADDR(rq->bd))>>8);	/* tell device where is the descriptor ring */
	csr32w(ctlr, Rqst, (PCIWADDR(rq->rsp))>>4);	/* tell device where is the status buffer */
	csr32w(ctlr, Rqconfig, 0x80811004);
#ifdef not
	csr32w(ctlr, Rqconfig, (2<<30) |			/* DMA on */
		(8<<20) |							/* nd = 2^8 */
		(1<<16) |							/* bsize = 8K */
		(1<<12) |							/* IRQ destination: normal */
		(0x10<<4));						/* timeout for closing Rx buffer: 0x10 * 32us */
#endif
	csr32w(ctlr, Rqhwp, rq->wp);				/* reset hwp */
	coherence();
	dprint("-- rqreset end rp %d wp %d hrp %d\n", rq->rp, rq->wp, *rq->rsp & 0xFFF);
	return 1;
}

/* should be called with the nic grabbed */
static int
rqstop(Ctlr *ctlr)
{
	dprint("-- rqstop start\n");
	csr32w(ctlr, Rqconfig, 0);					/* disable DMA */
	if (!csr32poll(ctlr, Rqstatus, 1<<24, 1<<24, 1000)) {	/* wait until DMA stopped */
		print("can't stop Rx DMA: %x\n", csr32r(ctlr, Rqstatus));
		return 0;
	}

	dprint("-- rstop end\n");
	return 1;
}

static int
rqavail(void *a)
{
	Rq *rq;

	rq = a;
	dprint("rqavail: rp %d rsp %d\n", rq->rp, (*rq->rsp) & 0xFFF);
	return rq->rp != ((*rq->rsp) & 0xFFF);
}

static Block *
rqbread(Ctlr *ctlr)
{
	uchar c;
	uint len;
	ushort seq;
	Block *b;
	Rq *rq;

	rq = &ctlr->rq;

again:
	sleep(&rq->rendez, rqavail, rq);
	qlock(rq);
	if (!rqavail(rq)) {
		qunlock(rq);
		goto again;
	}

	b = rq->blocks[rq->rp];
	rq->blocks[rq->rp] = nil;
	if (rqreplenish(rq)) {
		/* do we need to grab the nic??? */
		csr32w(ctlr, Rqhwp, rq->wp);		/* tell device the new wp */
	}
	qunlock(rq);

	coherence();
	len = (*(uint *)b->rp) & 0xFFFF;
	b->wp += len + 4;
	c = b->rp[4];
	seq = *(ushort *) (b->rp + 6);
	dprint("rqbread: block len %ux cmd %ux seq %ux\n", len, c, seq);
	if (!(seq & 0x8000) && c!=0xc0 && c!=0xc3 && c!=0xc5 && c!=0x9d && c!=0x1c)
		seq = tqframefree(ctlr, seq);

	dprint("rqbread: real seq: %ux\n", seq);
//	print("rp %d wp %d rsp %d\n", rq->rp, rq->wp, *rq->rsp);
	*(ushort *) (b->rp + 6) = seq;
	rq->rp++;
	if (rq->rp >= rq->nd)
		rq->rp = 0;

	return b;
}

Dev iwldevtab = {
	'w',
	"iwl",
	iwlreset,
	iwlinit,
	devshutdown,
	iwlattach,
	iwlwalk,
	iwlstat,
	iwlopen,
	devcreate,
	iwlclose,
	iwlread,
	iwlbread,
	iwlwrite,
	iwlbwrite,
	devremove,
	devwstat,
};




