#include <u.h>
#include <libc.h>
#include <thread.h>
#include "iwl.h"

#define MSEQ(id, tag, ignresp)	((((ignresp)?0x1f:(id))<<10) | (tag))
#define MID(seq)	((seq)>>10)
#define MIGN(seq)	(MID(seq) == 0x1f)
#define MTAG(seq)	((seq) & 0x3FF)

static void qtxproc(void *a);
static void qrxproc(void *a);

int
queueinit(int id)
{
	char *path;
	Iqueue *q;

	q = &iwl->queues[id];
	memset(q, 0, sizeof(*q));
	q->id = id;
	path = emalloc(strlen(iwl->path) + 16);
	sprint(path, "%s/tx%d", iwl->path, id);
	q->txfd = open(path, OWRITE);
	free(path);
	if (q->txfd < 0)
		return 0;

	q->txrend.l = &q->QLock;
	q->rxrend.l = &q->QLock;

	/* start a proc for writing to this queue */
	proccreate(qtxproc, q, 32768);
	return 1;
}

int
rxinit(void)
{
	char *path;

	path = emalloc(strlen(iwl->path) + 16);
	sprint(path, "%s/rx", iwl->path);
	iwl->rxfd = open(path, OREAD);
	free(path);
	if (iwl->rxfd < 0)
		return 0;

	proccreate(qrxproc, nil, 32768);
	return 1;
}

/* called with the lock held */
static ushort
tagget(Iqueue *q)
{
	int i, j;
	ushort tag;

	tag = q->nexttag++;
	q->tagmap[tag/32] |= 1 << (tag % 32);
	if (q->nexttag > 0x3FF)
		q->nexttag = 0;

	i = q->nexttag / 32;
	j = q->nexttag % 32;
	for(; i < nelem(q->tagmap); i++) {
		if (q->tagmap[i] != 0xFFFFFFFF)
			break;
		j = 0;
	}

	if (i >= nelem(q->tagmap))
		sysfatal("all tags used up");

	for(; j < 32; j++)
		if ((q->tagmap[i] & (1<<j)) == 0)
			break;

	q->nexttag = i*32 + j;

	return tag;
}

/* called with the lock held */
static void
tagput(Iqueue *q, ushort tag)
{
	q->tagmap[tag/32] &= ~(1 << (tag%32));
}

Imsg *
msgalloc(int sz)
{
	Imsg *m;

	m = emallocz(sizeof(*m) + sz + 8, 1);
	m->len = sz;
	m->pkt = (uchar *) &m[1];
	m->data = m->pkt + 8;

	return m;
}

void
msgfree(Imsg *m)
{
	free(m);
}

ushort
msgsend(int qid, Imsg *m, int ignresp)
{
	ushort seq;
	Iqueue *q;

	q = &iwl->queues[qid];
	qlock(q);
	seq = MSEQ(qid, tagget(q), ignresp);
	*((ushort *) &m->pkt[0]) = m->len + 4;
	*((ushort *) &m->pkt[2]) = m->datalen;
	m->pkt[4] = m->cmd;
	m->pkt[5] = 0;
	*((ushort *) &m->pkt[6]) = seq;
	if (q->txlast)
		q->txlast->next = m;
	else
		q->txfirst = m;
	q->txlast = m;
	rwakeup(&q->txrend);
	qunlock(q);

	return seq;
}

Imsg *
msgwait(int qid, int (*fn)(int, Imsg *, void *), void *a)
{
	Iqueue *q;
	Imsg *m, *pm;

	q = &iwl->queues[qid];
	qlock(q);
	while (1) {
		for(pm = nil, m = q->rxfirst; m != nil; pm = m, m = m->next) {
//			print("msgwait cmd %x seq %x\n", m->cmd, m->seq);
			if ((*fn)(qid, m, a))
				goto found;
		}

		if (pm) {
			/* the message is not for us, wakeup somebody else */
			rwakeup(&q->rxrend);
		}

//		fprint(2, "msgwait: wait for queue %p\n", q);
		rsleep(&q->rxrend);
	}

found:
	if (!(m->seq & 0x8000))
		tagput(q, MTAG(m->seq));

	if (pm)
		pm->next = m->next;
	else
		q->rxfirst = m->next;

	if (m->next == nil)
		q->rxlast = pm;

	qunlock(q);
	return m;
}

static void
qtxproc(void *a)
{
	int n;
	char buf[16];
	Imsg *m;
	Iqueue *q;

	q = a;
	snprint(buf, sizeof(buf), "iwltxproc%d", q->id);
	threadsetname(buf);
	qlock(q);
	while (1) {
		if (!q->txfirst) {
			rsleep(&q->txrend);
			continue;
		}

		m = q->txfirst;
		q->txfirst = m->next;
		if (m->next == nil)
			q->txlast = nil;
		qunlock(q);

		fprint(2, "qtxproc: send cmd %ux len %d\n", m->cmd, m->len);
		n = write(q->txfd, m->pkt, m->len + 8);
		if (n != m->len+8)
			sysfatal("short write");

		msgfree(m);
		qlock(q);
	}
}

static void
qrxproc(void *a)
{
	int i, n, blen, qid, ign;
	Imsg *m, *m1;
	Iqueue *q;

	USED(a);
	threadsetname("iwlrxproc");
	blen = 8192;
	m = msgalloc(blen);
	n = 0;
	while ((i = read(iwl->rxfd, m->pkt + n, blen - n)) > 0) {
//		print("cmdrproc: got %d bytes\n", i);
		n += i;
		if (n < 4)
			continue;

		m->len = 4 + *((uint *) m->pkt) & 0xFFFF;
		if (m->len > blen)
			sysfatal("message too big: %ux\n", m->len);

		if (n < m->len)
			continue;

		m->cmd = m->pkt[4];
		m->flags = m->pkt[5];
		m->seq = *((ushort *) &m->pkt[6]);
		m->data = &m->pkt[8];

		n -= m->len;
		m1 = msgalloc(blen);
		if (n > 0)
			memmove(m1->pkt, m->pkt + m->len, n);

		m->len -= 8;
		m->next = nil;

		if (m->cmd==0xc0 || m->cmd==0xc3 || m->cmd==0xc5 ||
				m->cmd==0x9d || m->cmd==0x1c)
			m->seq |= 0x8000;

		fprint(2, "qrxproc: cmd %ux seq %ux flags %ux\n", m->cmd, m->seq, m->flags);
		if (m->seq & 0x8000) {
			/* all messages generated from ucode should go
			     to the command queue and can't be ignored */
			qid = 4;
			ign = 0;
		} else {
			qid = MID(m->seq);
			ign = MIGN(m->seq);
		}

//		fprint(2, "qrxproc qid %d ign %d\n", qid, ign);
		if (!ign) {
			q = &iwl->queues[qid];
			qlock(q);
			if (q->rxlast)
				q->rxlast->next = m;
			q->rxlast = m;
				if (!q->rxfirst)
			q->rxfirst = m;
//			fprint(2, "qrxproc: wakeup queue %p\n", q);
			rwakeup(&q->rxrend);
			qunlock(q);
		} else {
			/* we allow ignoring responses only for the command queue */
			if (!(m->seq & 0x8000))
				tagput(&iwl->queues[4], MTAG(m->seq));

			free(m);
		}

		m = m1;
	}

	fprint(2, "cmdrproc: exit %d\n", i);

}