#include <u.h>
#include <libc.h>
#include <thread.h>
#include <iwl.h>

enum {			/* commands */
	CWimaxcoex	= 0x5a,		/* WIMAX co-existance */
	CCalibconf	= 0x65,		/* Calib config */
	CCalibresult	= 0x66,		/* Calib result */
	CCalibend		= 0x67,		/* Calibration complete */
	CBtcoex		= 0x9b,		/* Bluetooth co-existance */
	CCalib		= 0xb0,		/* PHY calibration */
};

enum {
	EExtal		= (2*0x128) | (EEtblcal<<16),
};

static int
isalive(int n, Imsg *m, void *a)
{
	USED(a);
	USED(n);

//	fprint(2, "isalive cmd %ux seq %ux\n", m->cmd, m->seq);
	return m->cmd == 0x01;
}

static int
cmdwimaxcoex(void)
{
	Imsg *m;

	m = msgalloc(68);
	m->cmd = CWimaxcoex;
	m = cmdrpc(m);
	fprint(2, "received wimax coex response\n");
	free(m);
	return 1;
}

static int
cmdbtcoex(void)
{
	Imsg *m;

	m = msgalloc(12);
	m->cmd = CBtcoex;
	m->data[0] = 3;				/* flags */
	m->data[1] = 0xaa;				/* lead time */
	m->data[2] = 1;				/* maxkill */
	m->data[3] = 0; 				/* reserved */
	*((uint *) &m->data[4]) = 0;		/* kill ack mask */
	*((uint *) &m->data[8]) = 0;		/* kill cts mask */

	m = cmdrpc(m);
	fprint(2, "received bt coex response\n");
	free(m);
	return 1;
}

static int
cmdxtalcalib(void)
{
	int xcaddr;
	Imsg *m;

	xcaddr = eeaddr(EExtal);
	m = msgalloc(6);
	m->cmd = CCalib;
	m->data[0] = 15;				/* calibrate crystal freq */
	m->data[4] = ee16r(xcaddr);
	m->data[5] = ee16r(xcaddr);
	m = cmdrpc(m);
	fprint(2, "received xtal calib response\n");
	free(m);
	return 1;
}

static int
checkcalib(int qid, Imsg *m, void *a)
{
	USED(qid);
	USED(a);
//	print("checkcalib: cmd %x\n", m->cmd);
	return m->cmd==CCalibconf ||
		m->cmd==CCalibresult ||
		m->cmd==CCalibend;
}

static int
cmdcalibconfig(Imsg **pm, int mlen)
{
	int n;
	uchar op;
	ushort seq;
	uint *p;
	Imsg *m, *mm;

	memset(pm, 0, mlen*sizeof(*pm));
	m = msgalloc(92);
	m->cmd = CCalibconf;
	p = (uint *) m->data;
	p[0] = ~0;					/* ucode/once: enable */
	p[1] = ~0;					/* ucode/once: start */
	p[2] = ~0;					/* ucode/once: sendres */
	p[10] = ~0;				/* ucode: flags */

	seq = cmdsend(m, 0);
	while (1) {
		m = cmdwait(checkcalib, &seq);
		print("cmd %ux seq %ux len %ux\n", m->cmd, m->seq, m->len);
//		for(i = 0; i < m->len + 8; i++)
//			print("%ux ", m->pkt[i]);
//		print("\n");
		if (m->cmd == CCalibend)
			break;
		else if (m->cmd == CCalibconf)
			continue;

		op = m->data[0];
		switch (op) {
		case 9:				/* calibrate LO */
			n = 0;
			break;
		case 11:				/* calibrate TX IQ */
			n = 1;
			break;
		case 17:				/* calibrate TX IQ PERD */
			n = 2;
			break;
		default:
			n = -1;
			break;
		}

		if (n < 0) {
			fprint(2, "unknown calibration opcode: %ux\n", op);
			free(m);
			continue;
		}

		if (mlen < n)
			sysfatal("m array too small");

		mm = msgalloc(m->len);
		mm->cmd = CCalib;
		memmove(mm->data, m->data, m->len);
		free(m);

		pm[n] = mm;
	}

	return 1;
}


int
iwlreset(void)
{
	int i;
	Imsg *m, *mcalib[3];

	if (!ucloadinit()) {
		fprint(2, "firmware upload failed\n");
		return 0;
	}

	if (!ctlsend("reset\n")) {
		fprint(2, "reset failed\n");
		return 0;
	}

	m = cmdwait(isalive, nil);
	if (m) {
		fprint(2, "received alive command\n");
		free(m);
	}

	if (!cmdwimaxcoex()) {
		fprint(2, "wimax coex failed\n");
		return 0;
	}

	if (!cmdxtalcalib()) {
		fprint(2, "xtal calibration failed\n");
		return 0;
	}

	if (!cmdcalibconfig(mcalib, nelem(mcalib))) {
		fprint(2, "calib config failed\n");
		return 0;
	}

	if (!ucloadruntime()) {
		fprint(2, "firmware upload failed\n");
		return 0;
	}

	if (!ctlsend("reset\n")) {
		fprint(2, "reset failed\n");
		return 0;
	}

	m = cmdwait(isalive, nil);
	if (m) {
		fprint(2, "received alive command\n");
		free(m);
	}

	if (!cmdwimaxcoex()) {
		fprint(2, "wimax coex failed\n");
		return 0;
	}

	if (!cmdxtalcalib()) {
		fprint(2, "xtal calibration failed\n");
		return 0;
	}

	for(i = 0; i < nelem(mcalib); i++) {
		if (mcalib[i] == nil)
			continue;

		m = cmdrpc(mcalib[i]);
		if (!m) {
			print("error while calibrating\n");
			threadexitsall("fail");
		}
		free(m);
	}

	if (!cmdbtcoex()) {
		fprint(2, "bt coex failed\n");
		return 0;
	}

	return 1;
}
